import aiohttp_jinja2
from riego.db import get_db
from riego.web.security import raise_permission


class Dashboard():
    def __init__(self, app):
        self._db_conn = get_db().conn

    @aiohttp_jinja2.template('dashboard/index.html')
    async def index(self, request):
        await raise_permission(request, permission=None)

        cursor = self._db_conn.cursor()
        cursor.execute("""SELECT *, date(last_run) AS date_last_run
                    FROM valves
                    WHERE is_hidden = 0""")
        items = cursor.fetchall()
        self._db_conn.commit()

        return {'valves': items}



from datetime import datetime
import re
import json
from sqlite3 import IntegrityError, Row
from typing import Any

from logging import getLogger
_log = getLogger(__name__)

_instance = None


def get_boxes():
    global _instance
    return _instance


def setup_boxes(options=None, db=None, mqtt=None):
    global _instance
    if _instance is not None:
        del _instance
    _instance = Boxes(options=options, db=db, mqtt=mqtt)
    return _instance


class Boxes():
    def __init__(self, options=None, db=None, mqtt=None):
        global _instance
        if _instance is None:
            _instance = self

        self._db_conn = db.conn
        self._mqtt = mqtt
        self._options = options

        self._mqtt.subscribe(self._options.mqtt_lwt_subscription,
                             self._mqtt_lwt_handler)
        self._mqtt.subscribe(self._options.mqtt_state_subscription,
                             self._mqtt_state_handler)
        self._mqtt.subscribe(self._options.mqtt_info1_subscription,
                             self._mqtt_info1_handler)
        self._mqtt.subscribe(self._options.mqtt_info2_subscription,
                             self._mqtt_info2_handler)

    async def _mqtt_lwt_handler(self, topic: str, payload: str) -> bool:
        """Create a new box or update an existing box
        This message comes from Tasmota only on startup and if
        we send a message for delete this reatained message from mqtt broker

        :param topic: topic of mqtt message, "tele/+/LWT"
        :type topic: str
        :param payload: [description]
        :type payload: str
        :return: [description]
        :rtype: bool
        """
        _log.debug(f'LWT: {topic}, payload: {payload}')
        if payload is None or payload == '':
            # Could be a retain delete message
            return

        box_topic = re.search('/(.*?)/', topic)
        if box_topic is None:
            return False
        box_topic = box_topic.group(1)

        first_seen = datetime.now()
        if payload == "Online":
            online_since = first_seen
        if payload == "Offline":
            online_since = None
            # TODO set all "Valves" to "offline"

        try:
            with self._db_conn:
                self._db_conn.execute(
                    ''' INSERT INTO boxes
                    (name, topic, first_seen, online_since)
                    VALUES (?, ?, ?, ?) ''',
                    (box_topic, box_topic, first_seen, online_since))
        except IntegrityError:
            with self._db_conn:
                self._db_conn.execute(
                    'UPDATE boxes SET online_since = ?  WHERE topic = ?',
                    (online_since, box_topic))
        return True

    async def _mqtt_state_handler(self, topic: str, payload: str) -> bool:
        """Insert lines into valves table for every Channel found in
        "/tele/+/STATE" message

        This message comes from Tasmota on startup and in intervalls

        :param topic: topic from mqtt message, "/tele/+/STATE"
        :type topic: str
        :param payload: payload sfrom mqtt message
        :type payload: str
        :return: True on success
        :rtype: bool
        """
        _log.debug(f'State: {topic}, payload: {payload}')
        if payload is None or payload == '':
            # Could be a config message
            return False

        box_id, box_topic = await self._get_box_id_by_topic(topic=topic)

        payload = json.loads(payload)
        for item in payload:
            # TODO if item "Wifi" take the Wifi nested dict and,
            #  search and extract "signal"
            channel_nr = re.match('^POWER(\d+)', item)  # noqa: W605
            if channel_nr is not None:
                channel_nr = channel_nr.group(1)
                try:
                    with self._db_conn:
                        self._db_conn.execute(
                            """INSERT INTO valves (channel_nr, box_id, name)
                            VALUES (?,?,?)""",
                            (channel_nr, box_id,
                             f'{box_topic}, Power {channel_nr}')
                        )
                except IntegrityError:
                    # _log.debug(f'unable to Update MQTT STATE:{e}')
                    pass
        return True

    async def _mqtt_info1_handler(self, topic: str, payload: str) -> bool:
        """Update existing box with additional info
        This message comes from Tasmota only on startup

        :param topic: topic from mqtt message, "/tele/+/INFO1"
        :type topic: str
        :param payload: payload from mqtt mesage
        :type payload: str
        :return: [description]
        :rtype: bool
        """
        _log.debug(f'Info1: {topic}, payload: {payload}')
        if payload is None or payload == '':
            # Could be a config message
            return False

        box_id, box_topic = await self._get_box_id_by_topic(topic=topic)

        payload = json.loads(payload)
        # TODO hw_type splitten in hw_type und hw_version
        hw_type = payload.get('Module', '')
        hw_version = ''
        # TODO sw_version splitten in sw_type und sw_version
        sw_version = payload.get('Version', '')
        sw_type = ''
        fallback_topic = payload.get('FallbackTopic', '')
        group_topic = payload.get('GroupTopic', '')

        try:
            with self._db_conn:
                self._db_conn.execute(
                    """UPDATE boxes
                    SET hw_type = ?, hw_version = ?,
                    sw_type= ?, sw_version = ?,
                    fallback_topic = ?, group_topic = ?
                    WHERE id = ?""",
                    (hw_type, hw_version, sw_type, sw_version,
                     fallback_topic, group_topic, box_id))
        except IntegrityError:
            # _log.debug(f'unable to Update MQTT INFO1:{e}')
            pass
        return True

    async def _mqtt_info2_handler(self, topic: str, payload: str) -> bool:
        """Update existing box with additional info
        This message comes from Tasmota only on startup

        :param topic: topic from mqtt message, "/tele/+/INFO2"
        :type topic: str
        :param payload: payload from mqtt message
        :type payload: str
        :return: [description]
        :rtype: bool
        """
        _log.debug(f'Info2: {topic}, payload: {payload}')
        if payload is None or payload == '':
            # Could be a config message
            return False

        box_id, box_topic = await self._get_box_id_by_topic(topic=topic)

        payload = json.loads(payload)
        hostname = payload.get('Hostname', '')
        ip_address = payload.get('IPAddress', '')

        try:
            with self._db_conn:
                self._db_conn.execute(
                    """UPDATE boxes
                    SET hostname = ?, ip_address = ?
                    WHERE id = ?""",
                    (hostname, ip_address, box_id))
        except IntegrityError:
            # _log.debug(f'Unable to update MQTT INFO2:{e}')
            pass
        return True

    async def _get_box_id_by_topic(self, topic=None):
        box_topic = re.search('/(.*?)/', topic)
        if box_topic is None:
            return False
        box_topic = box_topic.group(1)

        cursor = self._db_conn.cursor()
        cursor.execute('SELECT * FROM boxes WHERE topic=?', (box_topic,))
        item = cursor.fetchone()
        self._db_conn.commit()
        if item is None:
            with self._db_conn:
                cursor = self._db_conn.execute(
                    ''' INSERT INTO boxes
                    (name, topic, first_seen, online_since)
                    VALUES (?, ?, ?, ?) ''',
                    (box_topic, box_topic, datetime.now(), datetime.now()))
            box_id = cursor.lastrowid
        else:
            box_id = item['id']
        return box_id, box_topic

# Section with unused functions
#
#

    async def insert(self, item: dict) -> bool:
        try:
            with self._db_conn:
                cursor = self._db_conn.execute(
                    ''' INSERT INTO boxes
                    (topic, display_name, remark)
                    VALUES (?, ?, ?) ''',
                    (item['topic'], item['display_name'], item['remark']))
        except IntegrityError:
            ret = None
        else:
            ret = cursor.lastrowid
        return ret

    async def update(self, item_id: int, item: dict) -> bool:
        ret = True
        try:
            with self._db_conn:
                self._db_conn.execute(
                    ''' UPDATE boxes
                        SET display_name = ?, remark = ?
                        WHERE id = ? ''',
                    (item['display_name'], item['remark'], item_id))
        except IntegrityError:
            ret = False
        return ret

    async def delete(self, item_id: int) -> None:
        try:
            with self._db_conn:
                self._db_conn.execute(
                    'DELETE FROM boxes WHERE id = ?',
                    (item_id,))
        except IntegrityError:
            pass
        return None

    async def fetch_one_by_key(self, key: str, value: Any) -> Row:
        c = self._db_conn.cursor()
        sql = f'SELECT * FROM boxes WHERE {key}=?'
        c.execute(sql, (value,))
        ret = c.fetchone()
        self._db_conn.commit()
        return ret

    async def fetch_one_by_id(self, item_id: int) -> Row:
        c = self._db_conn.cursor()
        c.execute('SELECT * FROM boxes WHERE id=?', (item_id,))
        ret = c.fetchone()
        self._db_conn.commit()
        return ret

    async def fetch_all(self) -> Row:
        c = self._db_conn.cursor()
        c.execute('SELECT * FROM boxes')
        ret = c.fetchall()
        self._db_conn.commit()
        return ret



import asyncssh
import asyncio
import aiohttp


from logging import getLogger
_log = getLogger(__name__)

_instance = None


def get_ssh():
    global _instance
    return _instance


def setup_ssh(app=None, options=None, parameters=None):
    global _instance
    if _instance is not None:
        del _instance
    _instance = Ssh(app=app, options=options, parameters=parameters)
    return _instance


class Ssh:
    def __init__(self, app=None, options=None, parameters=None):
        global _instance
        if _instance is None:
            _instance = self

        self._parameters = parameters
        self._options = options
        self._STOP = False
        self._conn = None
        self._listener = None
        app.cleanup_ctx.append(self._ssh_engine)

    async def _ssh_engine(self, app):
        task = asyncio.create_task(self._startup(app))
        yield
        await self._shutdown(app, task)

    async def _startup(self, app) -> None:
        _log.debug('Ssh Engine startup called')
        await self.store_new_keys()

        while not self._STOP:
            if (self._parameters.ssh_server_hostname is None or
                self._parameters.ssh_server_port is None or
                self._parameters.ssh_server_listen_port is None or
                    self._parameters.ssh_user_key is None):
                await asyncio.sleep(3)
                continue
            ssh_user_key = asyncssh.import_private_key(
                self._parameters.ssh_user_key)
            async with asyncssh.connect(
                    self._parameters.ssh_server_hostname,
                    port=self._parameters.ssh_server_port,
                    username=self._parameters.cloud_identifier,
                    client_keys=ssh_user_key,
                    known_hosts=self._options.ssh_known_hosts) as self._conn:

                self._listener = await self._conn.forward_remote_port(
                    'localhost',  # Bind to localhost on remote Server
                    self._parameters.ssh_server_listen_port,
                    'localhost',
                    self._options.http_server_bind_port)
                await self._listener.wait_closed()
            await asyncio.sleep(3)

    async def _shutdown(self, app, task) -> None:
        _log.debug('Ssh Engine shutdown called')
        self._STOP = True
        if self._listener is not None:
            self._listener.close()
        if self._conn is not None:
            self._conn.close()
#            await self._conn.wait_closed()
        return None

    async def store_new_keys(self):
        key = asyncssh.generate_private_key(self._options.ssh_key_algorithm)
        self._parameters.ssh_user_key = key.export_private_key()
        public_user_key = key.export_public_key().decode('ascii')

        data = {'cloud_identifier': self._parameters.cloud_identifier,
                'public_user_key': public_user_key}

        async with aiohttp.ClientSession() as session:
            async with session.post(self._options.cloud_api_url,
                                    json=data) as resp:
                if resp.status != 200:
                    return
                data = await resp.json()
        self._parameters.ssh_server_hostname = data['ssh_server_hostname']  # noqa: E501
        self._parameters.ssh_server_port = data['ssh_server_port']
        self._parameters.ssh_server_listen_port = data['ssh_server_listen_port']  # noqa: E501



"""

"""

from yoyo import step

__depends__ = {'20210131_01_QndU0'}

steps = [
    step(
    '''CREATE TABLE "users" (
	"id"	        INTEGER,
	"identity"	    TEXT,
	"password"	    TEXT,
    "full_name"     TEXT,
    "email"         TEXT,
    "permission_id"	INTEGER,
	"is_superuser"	INTEGER DEFAULT 0,
	"is_disabled"	INTEGER DEFAULT 0,
	"remember_me"	TEXT,
	"created_at"	timestamp DEFAULT CURRENT_TIMESTAMP,
    CONSTRAINT "identity_uc" UNIQUE("identity"),
	PRIMARY KEY("id"))''',
    '''DROP TABLE users'''
    ),
    step(
    '''INSERT INTO "users"
    ("identity","password","is_superuser")
    VALUES ("admin","$2b$12$SsmDaUnej3koYln39Dq9Ue2VBjYd.FyGMeAV9kK3edRjAzLztIaCC",1)''',
      '''DELETE FROM users WHERE identity = "admin" '''
    ),
    step(
    ''' CREATE TABLE "users_permissions" (
	"id"	        INTEGER,
	"name"	        TEXT,
	"user_id"	    INTEGER NOT NULL REFERENCES "users"("id") ON DELETE CASCADE,
	"created_at"	timestamp DEFAULT CURRENT_TIMESTAMP,
	PRIMARY KEY("id"))''',
    ''' DROP TABLE "users_permissions" '''
    ),
    step(
    ''' CREATE TABLE "users_tokens" (
	"id"	        INTEGER,
	"sequence"	    TEXT NOT NULL,
    "hash"	        TEXT,
    "category"	    TEXT,
	"user_id"	    INTEGER NOT NULL REFERENCES "users"("id") ON DELETE CASCADE,
	"created_at"	timestamp DEFAULT CURRENT_TIMESTAMP,
    CONSTRAINT "sequence_uc" UNIQUE("sequence"),
	PRIMARY KEY("id"))''',
    ''' DROP TABLE "users_tokens" '''
    )


]



from sqlite3 import IntegrityError
import secrets


from logging import getLogger
_log = getLogger(__name__)

_instance = None


def get_parameters():
    global _instance
    return _instance


def setup_parameters(app=None, options=None, db=None):
    global _instance
    if _instance is not None:
        del _instance
    _instance = Parameters(app=app, options=options, db=db)
    return _instance


class Parameters:
    def __init__(self, app=None, options=None, db=None):
        global _instance
        if _instance is None:
            _instance = self
        self._db_conn = db.conn

        self._start_time_1 = options.parameters_start_time_1
        self._max_duration = options.parameters_max_duartion
        self._smtp_hostname = options.parameters_smtp_hostname
        self._smtp_port = options.parameters_smtp_port
        self._smtp_security = options.parameters_smtp_security
        self._smtp_user = options.parameters_smtp_user
        self._smtp_password = options.parameters_smtp_password
        self._smtp_from = options.parameters_smtp_from

        self._cloud_identifier = None

        self._ssh_server_hostname = None
        self._ssh_server_port = None
        self._ssh_server_listen_port = None
        self._ssh_user_key = None

        self._load_all()

    @property
    def start_time_1(self):
        return self._start_time_1

    @start_time_1.setter
    def start_time_1(self, value):
        self._start_time_1 = value
        self._update_value_by_key(key="start_time_1", value=value)

    @property
    def max_duration(self):
        return self._max_duration

    @max_duration.setter
    def max_duration(self, value):
        self._max_duration = value
        self._update_value_by_key(key="max_duration", value=value)

    @property
    def smtp_hostname(self):
        return self._smtp_hostname

    @smtp_hostname.setter
    def smtp_hostname(self, value):
        self._smtp_hostname = value
        self._update_value_by_key(key="smtp_hostname", value=value)

    @property
    def smtp_port(self):
        return self._smtp_port

    @smtp_port.setter
    def smtp_port(self, value):
        self._smtp_port = value
        self._update_value_by_key(key="smtp_port", value=value)

    @property
    def smtp_security(self):
        return self._smtp_security

    @smtp_security.setter
    def smtp_security(self, value):
        self._smtp_security = value
        self._update_value_by_key(key="smtp_security", value=value)

    @property
    def smtp_user(self):
        return self._smtp_user

    @smtp_user.setter
    def smtp_user(self, value):
        self._smtp_user = value
        self._update_value_by_key(key="smtp_user", value=value)

    @property
    def smtp_password(self):
        return self._smtp_password

    @smtp_password.setter
    def smtp_password(self, value):
        self._smtp_password = value
        self._update_value_by_key(key="smtp_password", value=value)

    @property
    def smtp_from(self):
        return self._smtp_from

    @smtp_from.setter
    def smtp_from(self, value):
        self._smtp_from = value
        self._update_value_by_key(key="smtp_from", value=value)

    @property
    def ssh_server_hostname(self):
        return self._ssh_server_hostname

    @ssh_server_hostname.setter
    def ssh_server_hostname(self, value):
        self._ssh_server_hostname = value
        self._update_value_by_key(key="ssh_server_hostname", value=value)

    @property
    def ssh_server_port(self):
        return self._ssh_server_port

    @ssh_server_port.setter
    def ssh_server_port(self, value):
        self._ssh_server_port = value
        self._update_value_by_key(key="ssh_server_port", value=value)

    @property
    def ssh_server_listen_port(self):
        return self._ssh_server_listen_port

    @ssh_server_listen_port.setter
    def ssh_server_listen_port(self, value):
        self._ssh_server_listen_port = value
        self._update_value_by_key(key="ssh_server_listen_port", value=value)

    @property
    def ssh_user_key(self):
        return self._ssh_user_key

    @ssh_user_key.setter
    def ssh_user_key(self, value):
        self._ssh_user_key = value
        self._update_value_by_key(key="ssh_user_key", value=value)

    @property
    def cloud_identifier(self):
        if self._cloud_identifier is None:
            self._cloud_identifier = secrets.token_urlsafe(12)
            self._update_value_by_key(key="cloud_identifier",
                                      value=self._cloud_identifier)
        return self._cloud_identifier

    @cloud_identifier.setter
    def cloud_identifier(self, value):
        self._cloud_identifier = value
        self._update_value_by_key(key="cloud_identifier", value=value)

    def _update_value_by_key(self, key=None, value=None) -> bool:
        ret = True
        try:
            with self._db_conn:
                self._db_conn.execute(
                    'UPDATE parameters SET value = ? WHERE key = ?',
                    (value, key))
        except IntegrityError:
            ret = False
        try:
            with self._db_conn:
                self._db_conn.execute(
                    "INSERT INTO parameters (value, key) VALUES (?,?)",
                    (value, key))
        except IntegrityError:
            ret = False
        return ret

    def _load_all(self):
        c = self._db_conn.cursor()
        c.execute("SELECT * from parameters")
        items = c.fetchall()
        self._db_conn.commit()
        for item in items:
            attr_name = "_" + item['key']
            if getattr(self, attr_name, None) is not None:
                setattr(self, attr_name, item['value'])



from typing import Any, Dict
import aiohttp_jinja2
from aiohttp import web
from aiohttp_session import get_session

from sqlite3 import IntegrityError
from riego.db import get_db
from riego.web.security import (get_user, password_check,
                                password_hash, delete_websocket_auth)
import secrets
import asyncio


from logging import getLogger
_log = getLogger(__name__)

router = web.RouteTableDef()


def setup_routes_users(app):
    app.add_routes(router)


@router.get("/login", name='login')
@aiohttp_jinja2.template("users/login.html")
async def login(request: web.Request) -> Dict[str, Any]:
    redirect = request.rel_url.query.get("redirect", "")
    csrf_token = secrets.token_urlsafe()
#    session = await new_session(request)
    session = await get_session(request)
    session['csrf_token'] = csrf_token
    return {'csrf_token': csrf_token, 'redirect': redirect}


@ router.post("/login")
async def login_apply(request: web.Request) -> Dict[str, Any]:
    form = await request.post()
    session = await get_session(request)
    if session.get('csrf_token') != form['csrf_token']:
        # Normally not possible
        await asyncio.sleep(2)
        raise web.HTTPUnauthorized()

    if form.get('identity') is None:
        await asyncio.sleep(2)
        raise web.HTTPSeeOther(request.app.router['login'].url_for())

    cursor = get_db().conn.cursor()
    cursor.execute("""SELECT *, 'login' AS 'provider'
                    FROM users
                    WHERE identity = ?""", (form['identity'],))
    user = cursor.fetchone()

    if user is None or user['is_disabled'] or not len(user['password']):
        await asyncio.sleep(2)
        raise web.HTTPSeeOther(request.app.router['login'].url_for())
#    if not bcrypt.checkpw(form['password'].encode('utf-8'), user['password']):
    if not password_check(form['password'].encode('utf-8'), user['password']):
        await asyncio.sleep(2)
        raise web.HTTPSeeOther(request.app.router['login'].url_for())

    session['user_id'] = user['id']

    location = form.get('redirect')
    if location is None or location == '':
        location = request.app.router['home'].url_for()
    response = web.HTTPSeeOther(location=location)
# TODO use create_remember_me_auth from modul security
    if form.get('remember_me') is not None:
        remember_me = secrets.token_urlsafe()
        try:
            with get_db().conn:
                get_db().conn.execute(
                    '''UPDATE users
                       SET remember_me = ?
                       WHERE id = ? ''',
                    (remember_me, user['id']))
        except IntegrityError:
            pass
        response.set_cookie("remember_me", remember_me,
                            max_age=request.app['options'].max_age_remember_me,
                            httponly=True,
                            samesite='strict')
    return response


@ router.get("/logout", name='logout')
async def logout(request: web.Request) -> Dict[str, Any]:
    user = await get_user(request)
    if user is not None:
        await delete_websocket_auth(request, user=user)
# TODO use delete_remember_me_auth from modul security
    try:
        with get_db().conn:
            get_db().conn.execute("""UPDATE users
                                     SET remember_me = ''
                                     WHERE id = ?""", (user['id'],))
    except IntegrityError:
        pass
    session = await get_session(request)
    if session is not None:
        session.pop('user_id', None)
    response = web.HTTPSeeOther(request.app.router['login'].url_for())
#    response.set_cookie('remember_me', None,
#                        expires='Thu, 01 Jan 1970 00:00:00 GMT')
    response.del_cookie('remember_me')
    return response


@ router.get("/passwd", name='passwd')
@ aiohttp_jinja2.template("users/passwd.html")
async def passwd(request: web.Request) -> Dict[str, Any]:
    return {}


@ router.post("/passwd")
async def passwd_apply(request: web.Request) -> Dict[str, Any]:
    form = await request.post()
    user = await get_user(request)

# TODO check old_password and equality of pw1 an pw2
    password = form['new_password_1'].encode('utf8')
#    password = bcrypt.hashpw(password, bcrypt.gensalt(12))
    password = password_hash(password)
    try:
        with get_db().conn:
            get_db().conn.execute(
                '''UPDATE users
                    SET password = ?
                    WHERE id = ? ''', (password, user['id']))
    except IntegrityError:
        raise web.HTTPSeeOther(request.app.router['passwd'].url_for())

    raise web.HTTPSeeOther(request.app.router['home'].url_for())
    return {}  # not reached


@ router.get("/users", name='users')
@ aiohttp_jinja2.template("users/index.html")
async def index(request: web.Request) -> Dict[str, Any]:
    cursor = get_db().conn.cursor()
    cursor.execute('SELECT * FROM users')
    items = cursor.fetchall()
    get_db().conn.commit()
    return {"items": items}


@ router.get("/users/new", name='users_new')
@ aiohttp_jinja2.template("users/new.html")
async def new(request: web.Request) -> Dict[str, Any]:
    return {}


@ router.post("/users/new")
async def new_apply(request: web.Request) -> Dict[str, Any]:
    item = await request.post()
    try:
        with get_db().conn:
            cursor = get_db().conn.execute(
                ''' INSERT INTO users
                (identity, password,
                full_name, email,
                permission_id, is_superuser,
                is_disabled, remember_me)
                VALUES (?, ?, ?, ?, ?, ?) ''',
                (item['identity'], item['password'],
                 item['full_name'], item['email'],
                 item['permission_id'], item['is_superuser'],
                 item['is_disabled'], item['remember_me']))
    except IntegrityError as e:
        _log.debug(f'users add: {e}')
        raise web.HTTPSeeOther(request.app.router['users_new'].url_for())
    else:
        item_id = str(cursor.lastrowid)
        raise web.HTTPSeeOther(
            request.app.router['users_item_view'].url_for(item_id=item_id))
    return {}  # not reached


@ router.get("/users/{item_id}", name='users_item_view')
@ aiohttp_jinja2.template("users/view.html")
async def view(request: web.Request) -> Dict[str, Any]:
    item_id = request.match_info["item_id"]
    cursor = get_db().conn.cursor()
    cursor.execute('SELECT * FROM users WHERE id=?', (item_id,))
    item = cursor.fetchone()
    get_db().conn.commit()
    if item is None:
        raise web.HTTPSeeOther(request.app.router['users'].url_for())
    return {"item": item}


@ router.get("/users/{item_id}/edit", name='users_item_edit')
@ aiohttp_jinja2.template("users/edit.html")
async def edit(request: web.Request) -> Dict[str, Any]:
    item_id = request.match_info["item_id"]
    cursor = get_db().conn.cursor()
    cursor.execute('SELECT * FROM users WHERE id=?', (item_id,))
    item = cursor.fetchone()
    get_db().conn.commit()
    if item is None:
        raise web.HTTPSeeOther(request.app.router['users'].url_for())
    return {"item": item}


@ router.post("/users/{item_id}/edit")
async def edit_apply(request: web.Request) -> web.Response:
    item_id = request.match_info["item_id"]
    item = await request.post()
    try:
        with get_db().conn:
            get_db().conn.execute(
                '''UPDATE users
                   SET identity = ?, password = ?,
                   full_name = ?, email = ?,
                   permission_id = ?, is_superuser = ?,
                   is_disabled = ?, remember_me = ?
                   WHERE id = ? ''',
                (item['identity'], item['password'],
                 item['full_name'], item['email'],
                 item['permission_id'], item['is_superuser'],
                 item['is_disabled'], item['remember_me'],
                 item_id))
    except IntegrityError as e:
        _log.debug(f'box.view edit: {e}')
        raise web.HTTPSeeOther(
            request.app.router['users_item_edit'].url_for(item_id=item_id))
    else:
        raise web.HTTPSeeOther(
            request.app.router['users_item_view'].url_for(item_id=item_id))
    return {}  # Not reached


@ router.get("/users/{item_id}/delete", name='users_item_delete')
async def delete(request: web.Request) -> web.Response:
    item_id = request.match_info["item_id"]
    try:
        with get_db().conn:
            get_db().conn.execute(
                'DELETE FROM users WHERE id = ?',
                (item_id,))
    except IntegrityError as e:
        _log.debug(f'users  delete: {e}')
    raise web.HTTPSeeOther(request.app.router['users'].url_for())
    return {}  # Not reached



from datetime import datetime, timedelta
import asyncio
from sqlite3 import IntegrityError
import logging


_log = logging.getLogger(__name__)

_instance = None


def get_timer():
    global _instance
    return _instance


def setup_timer(app=None, options=None,
                db=None, mqtt=None,
                parameters=None, valves=None):
    global _instance
    if _instance is not None:
        del _instance
    _instance = Timer(app=app, options=options,
                      db=db, mqtt=mqtt,
                      parameters=parameters, valves=valves)
    return _instance


class Timer():
    def __init__(self, app=None, options=None,
                 db=None, mqtt=None,
                 parameters=None, valves=None):
        self._options = options
        self._db_conn = db.conn
        self._mqtt = mqtt
        self._parameters = parameters
        self._valves = valves

        self._running_period_end = None
        self._running_period_start = None

        self._stop = False
        self._task = None

        app.cleanup_ctx.append(self.timer_engine)

    async def timer_engine(self, app):
        self._task = asyncio.create_task(self._timer_loop())
        yield
        self._shutdown(self._task)

    async def _timer_loop(self) -> None:
        _log.debug('Timer Engine started')
        while not self._stop:
            await asyncio.sleep(3)
            await self._check_updates()
            c = self._db_conn.cursor()
            c.execute("""SELECT * FROM valves
                        ORDER BY valves.prio""")
            valves = c.fetchall()
            self._db_conn.commit()

            one_valve_is_on = 0
            for valve in valves:
                if valve['is_running'] == 1:
                    one_valve_is_on = 1
                    break

            for valve in valves:
                if await self._dispatch_valve(valve, one_valve_is_on):
                    break
        # Close last valve on exit
        if valve is not None:
            await self._valves.set_off_try(valve['id'])
        return None

    async def _check_updates(self):
        return None

    async def _dispatch_valve(self, valve, one_valve_is_on):
        # _log.debug("dispatch_valve: {}".format(valve['name']))
        if not self._mqtt.client.is_connected:
            return 0
        if valve is None:
            return 0
        if valve['duration'] == 0:
            return 0
        if valve['is_running'] == 1:
            return await self._check_to_switch_off(valve)
        if valve['is_hidden']:
            return 0
        if not valve['is_enabled']:
            return 0
        if valve['is_running'] == -1:
            return 0
        if valve['is_running'] == 0 and not one_valve_is_on == 1:
            return await self._check_to_switch_on(valve)
        return 0

    async def _check_to_switch_off(self, valve) -> bool:
        """
        If we had some action, we return True
        """
        _log.debug("check_to_switch_off for id: {}".format(valve['id']))
        if self._options.enable_timer_dev_mode:
            td = timedelta(minutes=0, seconds=valve['duration'])
        else:
            td = timedelta(minutes=valve['duration'])

        if datetime.now() - valve['last_run'] >= td:
            # Laufzeit erreicht
            await self._valves.set_off_try(valve['id'])
            _log.debug('valveOff: {}'.format(valve['name']))
            return True
        return False

    async def _check_to_switch_on(self, valve) -> bool:
        """
        If we had some action, we return True
        """
        _log.debug("check_to_switch_on for id: {}".format(valve['id']))
        if self._options.enable_timer_dev_mode:
            td = timedelta(days=0, seconds=valve['interval'])
        else:
            td = timedelta(days=valve['interval'])

        current_shedule_datetime = await self._is_running_period()
        if (datetime.now() - valve['last_shedule'] >= td and
                current_shedule_datetime is not None):
            # Intervall erreicht
            await self._valves.set_on_try(valve['id'])
            try:
                with self._db_conn:
                    self._db_conn.execute(
                        """UPDATE valves SET last_shedule = ? WHERE id = ? """,
                        (current_shedule_datetime, valve['id']))
            except IntegrityError as e:
                pass
                _log.error(f'update for last_shedule failed: {e}')
                return False
            _log.debug('valveOn: {}'.format(valve['name']))
            return True
        return False

    def _shutdown(self, task) -> None:
        _log.debug('Timer Engine shutdown called')
        self._stop = True

    async def _is_running_period(self) -> datetime:
        """Return current Shedule Datetime if we have raining times

        :return: [description]
        :rtype: datetime
        """
        max_duration = self._parameters.max_duration
        start_time_1 = self._parameters.start_time_1
        start_hour, start_minute = start_time_1.split(':')

        if self._running_period_start is None:
            self._running_period_start = datetime.now().replace(
                hour=int(start_hour), minute=int(start_minute),
                second=0, microsecond=0)

        if datetime.now() < self._running_period_start:
            self._running_period_end = None
            return None

        if self._running_period_end is None:
            self._running_period_end = self._running_period_start + \
                timedelta(minutes=int(max_duration))

        if datetime.now() > self._running_period_end:
            self._running_period_start = None
            return None
        return self._running_period_start



from datetime import datetime
import re
import json

from sqlite3 import IntegrityError

from logging import getLogger
_log = getLogger(__name__)


bool_to_int = {'true': 1, 'false': 0, True: 1, False: 0,
               'True': 1, 'False': 0, 'on': 1, 'off': 0,
               'On': 1, 'Off': 0, 'ON': 1, 'OFF': 0}

int_to_js_bool = {1: "true", 0: "false", -1: "-1"}


_instance = None


def get_valves():
    global _instance
    return _instance


def setup_valves(options=None, db=None, mqtt=None, websockets=None):
    global _instance
    if _instance is not None:
        del _instance
    _instance = Valves(options=options, db=db,
                       mqtt=mqtt, websockets=websockets)
    return _instance


class Valves():
    def __init__(self, options=None, db=None, mqtt=None, websockets=None):
        self._db_conn = db.conn
        self._mqtt = mqtt
        self._options = options
        self._websockets = websockets

        self.__is_running = None

        self._websockets.subscribe('valves', self._ws_handler)
        self._mqtt.subscribe(self._options.mqtt_result_subscription,
                             self._mqtt_result_handler)

    async def _ws_handler(self, msg: dict) -> None:
        _log.debug(f'In Valves._ws_handler: {msg}')
        if msg['action'] == 'update':
            await self.update_by_key(valve_id=msg['id'],
                                     key=msg['key'],
                                     value=msg['value'])
        return None

    async def _mqtt_result_handler(self, topic: str, payload: str) -> bool:
        """Dispatch mqtt messages:
        "stat/<box_topic>/RESULT {POWER1 :  ON}" (Is send after every cmnd)
        and store the state of channels to Database

        :param topic: Topic of subscribed MQTT-message
        :type topic: str
        :param payload: Payload of subscribed MQTT-message
        :type payload: str
        :return: [description]
        :rtype: bool
        """
        _log.debug(f'RESULT: {topic}, payload: {payload}')
        box_topic = re.search('/(.*?)/', topic)
        if box_topic is None:
            return False
        box_topic = box_topic.group(1)
        # convert payload to dict
        payload = json.loads(payload)
        # only one element in dict: {POWER1: ON}
        # extract channel_nr and status
        for key in payload:
            channel_nr = re.match('^POWER(\d+)', key)  # noqa: W605
            if channel_nr is None:
                return False
            channel_nr = channel_nr.group(1)
            status = payload[key]

            status = bool_to_int.get(status)
            if status == 1:
                await self._set_on_confirm(box_topic=box_topic,
                                           channel_nr=channel_nr)
            else:
                await self._set_off_confirm(box_topic=box_topic,
                                            channel_nr=channel_nr)
        return True

    async def _send_status_ws(self, valve_id=None, key=None, value=None):
        ret = {
            'action': "status",
            'scope': "valves",
            'id': valve_id,
            'key': key,
            'value': value,
        }
        ret = json.dumps(ret)
        await self._websockets.send_to_all(ret)
        return None

    async def _send_mqtt(self,
                         box_topic=None,
                         channel_nr=None,
                         payload=None) -> bool:
        if self._mqtt.client is None:
            return False
        if not self._mqtt.client.is_connected:
            return False

        topic = "{prefix}/{box_topic}/POWER{channel_nr}".format(
            prefix=self._options.mqtt_cmnd_prefix,
            box_topic=box_topic,
            channel_nr=channel_nr)
        self._mqtt.client.publish(topic, payload)
        return True

    async def set_on_try(self, valve_id) -> None:
        try:
            with self._db_conn:
                self._db_conn.execute(
                    "UPDATE valves SET is_running =? WHERE id = ?",
                    (-1, valve_id))
        except IntegrityError as e:
            _log.error(f"Unable to update: {e}")
            return None

        c = self._db_conn.cursor()
        c.execute("""SELECT valves.*, boxes.topic AS box_topic
                    FROM valves, boxes
                    WHERE valves.box_id = boxes.id AND valves.id=?""",
                  (valve_id,))
        valve = c.fetchone()
        self._db_conn.commit()
        if valve is None:
            _log.error("Valve not found 001")
            return None
        await self._send_status_ws(valve_id=valve_id,
                                   key='is_running',
                                   value=-1)
        await self._send_mqtt(box_topic=valve['box_topic'],
                              channel_nr=valve['channel_nr'],
                              payload=self._options.mqtt_keyword_ON)
        return None

    async def set_off_try(self, valve_id) -> None:
        try:
            with self._db_conn:
                self._db_conn.execute(
                    "UPDATE valves SET is_running =? WHERE id = ?",
                    (-1, valve_id))
        except IntegrityError as e:
            _log.error(f"Unable to update: {e}")
            return None

        c = self._db_conn.cursor()
        c.execute("""SELECT valves.*, boxes.topic AS box_topic
                    FROM valves, boxes
                    WHERE valves.box_id = boxes.id AND valves.id=?""",
                  (valve_id,))
        valve = c.fetchone()
        if valve is None:
            _log.error("Valve not found 001a")
            return None
        self._db_conn.commit()
        await self._send_status_ws(valve_id=valve_id,
                                   key='is_running',
                                   value=-1)
        await self._send_mqtt(box_topic=valve['box_topic'],
                              channel_nr=valve['channel_nr'],
                              payload=self._options.mqtt_keyword_OFF)
        return None

    async def _set_on_confirm(self, box_topic=None, channel_nr=None) -> None:
        c = self._db_conn.cursor()
        c.execute("""SELECT valves.*, boxes.topic AS box_topic
                    FROM valves, boxes
                    WHERE valves.box_id = boxes.id
                    AND boxes.topic =?
                    AND valves.channel_nr=?""", (box_topic, channel_nr))
        valve = c.fetchone()
        if valve is None:
            _log.error("Valve not found 002")
            return None
        self._db_conn.commit()
        try:
            with self._db_conn:
                self._db_conn.execute(
                    """UPDATE valves SET is_running = ?, last_run = ?
                    WHERE id = ?""",
                    (1, datetime.now(), valve['id']))
        except IntegrityError as e:
            _log.error(f"Unable to update: {e}")
            return None
        try:
            with self._db_conn:
                self._db_conn.execute(
                    "INSERT INTO events (valve_id) VALUES (?)",
                    (valve['id'],))
        except IntegrityError as e:
            _log.error(f'Unable to insert: {e}')

        await self._send_status_ws(valve_id=valve['id'],
                                   key='is_running',
                                   value=1)
        _log.info('Valve switched to ON: {}'.format(valve['name']))
        return None

    async def _set_off_confirm(self, box_topic=None, channel_nr=None) -> None:
        c = self._db_conn.cursor()
        c.execute("""SELECT valves.*, boxes.topic AS box_topic
                    FROM valves, boxes
                    WHERE valves.box_id = boxes.id
                    AND boxes.topic = ?
                    AND valves.channel_nr = ?""", (box_topic, channel_nr))
        valve = c.fetchone()
        self._db_conn.commit()
        if valve is None:
            _log.error("Valve not found 003")
            return None
        self._db_conn.commit()
        try:
            with self._db_conn:
                self._db_conn.execute(
                    "UPDATE valves SET is_running = ? WHERE id = ?",
                    (0, valve['id']))
        except IntegrityError as e:
            _log.error(f"Unable to update: {e}")
            return None

        c = self._db_conn.cursor()
        c.execute("""SELECT * FROM events
                    WHERE events.duration = 0
                    AND valve_id = ?
                    ORDER BY events.created_at DESC""", (valve['id'],))
        event = c.fetchone()
        if event is None:
            # Will happen after restart of Box
            _log.info("Event not found")
            return None
        self._db_conn.commit()
        # TODO sqlite should convert to datetime object
        duration = datetime.now() - event['created_at']
        duration = duration.total_seconds() / 60.0
        duration = round(duration)
        if duration == 0:
            duration = 1
        try:
            with self._db_conn:
                self._db_conn.execute(
                    "UPDATE events SET duration = ? WHERE id = ?",
                    (duration, event['id']))
        except IntegrityError as e:
            _log.error(f"Unable to update: {e}")
            return None

        await self._send_status_ws(valve_id=valve['id'], key='is_running',
                                   value=0)
        _log.info('Valve switched to OFF: {}'.format(valve['name']))
        return None

    async def update_by_key(self, valve_id=None, key=None, value=None) -> bool:
        ret = True
        if key == "is_running":
            if bool_to_int.get(value, 0):
                await self.set_on_try(valve_id)
            else:
                await self.set_off_try(valve_id)
            return True
        sql = f"UPDATE valves SET {key} = ? WHERE id = ?"
        try:
            with self._db_conn:
                self._db_conn.execute(sql, (value, valve_id))
        except IntegrityError as e:
            _log.error(f'Unable to update: {e}')
            ret = False
        print(f'valve_id={valve_id}, key={key}, value={value}')
        await self._send_status_ws(valve_id=valve_id, key=key, value=value)
        return ret


"""
   async def get_next(self, valve: Row) -> Row:
        valves = await self.fetch_all()
        count_valves = len(valves)
        if count_valves == 0:
            return None
        if valve is None:
            return valves[0]
        if count_valves == 1:
            return valves[0]
        for i in range(count_valves):
            if i == count_valves-1:
                return valves[0]
            if valve['id'] == valves[i]['id']:
                return valves[i+1]
        return None
"""
